import json

import pandas as pd

from pathlib import Path

from nova_utils.interfaces.dataset_iterable import DatasetIterable

import numpy as np


class HcaiAffectnetIterable(DatasetIterable):

    IMAGE_FOLDER_COL = "image_folder"

    LABELS = [
        "neutral",
        "happy",
        "sad",
        "suprise",
        "fear",
        "disgust",
        "anger",
        "contempt",
        "none",
        "uncertain",
        "non-face",
    ]

    def __init__(
            self,
            *args,
            dataset_dir=None,
            include_auto=False,
            ignore_duplicate=True,
            ignore_broken=True,
            ignore_unsupported_format=True,
            ignore_lists=None,
            **kwargs
    ):
        """
        Args:
          ignore_duplicate: bool. Flag to determine whether the duplicated files in the dataset should be included. Only affects the training set.
          ignore_broken: bool. Flag to determine whether the broken files in the dataset should be included.Affects all sets.
          ignore_wrong_format:  bool. Flag to determine whether files that are not in tensorflow compatible encoding should be ignored. Affects all sets.
          ignore_lists: list. Custom ignore lists for additional configurations.
          include_auto: bool. Flag to determine whether the automatically annotated files should be included in the dataset.
          **kwargs: keyword arguments forwarded to super.
        """
        super(HcaiAffectnetIterable, self).__init__(*args, **kwargs)
        if ignore_lists is None:
            ignore_lists = []

        if ignore_duplicate:
            ignore_lists.append("affect_net_ignore_list_duplicates.json")

        if ignore_broken:
            ignore_lists.append("affect_net_broken_files_ignore_list.json")

        if ignore_unsupported_format:
            ignore_lists.append("affect_net_ignore_list_wrong_format.json")

        self.include_auto = include_auto
        self.ignore_lists = ignore_lists

        self.dataset_dir = dataset_dir
        self._load_meta_data()

        if self.split == "train":
            self._iter = self._yield_examples(self.train_df)
        else:
            self._iter = self._yield_examples(self.test_df)

    def __iter__(self):
        return self._iter

    def __next__(self):
        return self._iter.__next__()

    def get_output_info(self):
        return {
            "index": {"dtype": np.str, "shape": (1,)},
            "image": {"dtype": np.str, "shape": (1,)},
            "expression": {"dtype": np.uint8, "shape": (1,)},
            "arousal": {"dtype": np.float32, "shape": (1,)},
            "valence": {"dtype": np.float32, "shape": (1,)},
            "facial_landmarks": {"dtype": np.uint32, "shape": (68, 2)},
            "rel_file_path": {"dtype": np.str, "shape": (1,)},
        }

    def _yield_examples(self, label_df):
        for index, row in label_df.iterrows():
            landmarks = np.fromstring(
                row["facial_landmarks"], sep=";", dtype=np.float32
            ).reshape((68, 2))

            yield {
                "index": index,
                "image": str(Path(self.dataset_dir) / row[self.IMAGE_FOLDER_COL] / index),
                "expression": row["expression"],
                "arousal": row["arousal"],
                "valence": row["valence"],
                "facial_landmarks": landmarks,
                "rel_file_path": row[self.IMAGE_FOLDER_COL] + "/" + index
            }

    def _load_meta_data(self):

        print("Loading Labels...")
        train_csv_path = (
                Path(self.dataset_dir) / "Manually_Annotated_file_lists" / "training.csv"
        )

        # The official test set has not been released in this version so we use the validation set as test set
        test_csv_path = (
                Path(self.dataset_dir) / "Manually_Annotated_file_lists" / "validation.csv"
        )

        train_csv_path_auto = (
                Path(self.dataset_dir)
                / "Automatically_Annotated_file_lists"
                / "automatically_annotated.csv"
        )

        train_df = pd.read_csv(train_csv_path, index_col=0)
        test_df = pd.read_csv(test_csv_path, index_col=0)
        train_df[self.IMAGE_FOLDER_COL] = ["Manually_Annotated_Images"] * len(train_df)
        test_df[self.IMAGE_FOLDER_COL] = ["Manually_Annotated_Images"] * len(test_df)

        # append automatic labels if not ignored:
        if self.include_auto:
            train_df_auto = pd.read_csv(train_csv_path_auto, index_col=0)
            train_df_auto[self.IMAGE_FOLDER_COL] = ["Automatically_Annotated_Images"]
            train_df = pd.concat([train_df, train_df_auto])

        len_train = len(train_df)
        len_test = len(test_df)
        print(
            "...loaded {} images for train\n...loaded {} images for test".format(
                len_train, len_test
            )
        )

        # removing labels that are specified in the ignore-lists
        print("Applying ignore lists...")

        filter_list_path = Path(__file__).parent / "Ignore_Lists"
        for filter_list in self.ignore_lists:
            print("... {}:".format(filter_list))
            with open(filter_list_path / filter_list) as json_file:
                filter = json.load(json_file)
                train_df.drop(filter, errors="ignore", inplace=True)
                test_df.drop(filter, errors="ignore", inplace=True)
            print(
                "\t...dropped {} images from train\n\t... dropped {} images from test".format(
                    len_train - len(train_df), len_test - len(test_df)
                )
            )
            len_train = len(train_df)
            len_test = len(test_df)

        print(
            "Final set sizes: \nTrain {}\nTest {}".format(len(train_df), len(test_df))
        )

        self.train_df = train_df
        self.test_df = test_df



from torch.utils.data import Dataset as TorchDataset
from torch.utils.data.dataset import T_co
import tensorflow as tf


class PyTorchDatasetWrapper(TorchDataset):
    """
    Adapts an non-batched, finite tensorflow dataset into a pytorch dataset for use with a pytorch dataloader.
    This class can process both dict and tuple output datasets.
    For performance reasons, the wrapper will batch the set internally,
    but will still return single data points as numpy arrays for the dataloader.
    """

    def __init__(self, dataset: tf.data.Dataset, internal_cache_size: int = 32):
        """
        :param dataset: any unbatched, finite tensorflow dataset
        :param internal_cache_size: the size of the internal batch cache
        """
        self._original_ds_reference = dataset
        self._output_dict = False
        self._iter = None
        self._batch_cache = None
        self._batch_cache_size = internal_cache_size
        self._internal_cache_size = 32
        self._configure()

    def _configure(self):

        # is the set dict or tuple based?
        if isinstance(self._original_ds_reference.element_spec, dict):
            self._output_dict = True

        self._original_ds_reference = self._original_ds_reference.batch(
            batch_size=self._internal_cache_size
        )

        # iteratively find total item count, since tf.cardinality is unreliably in this context
        self._len = 0
        self._first_member = 0
        for i, b in enumerate(self._original_ds_reference):
            if self._output_dict and i == 0:
                self._first_member = list(b.keys())[0]
            self._len = self._len + tf.shape(b[self._first_member])[0].numpy()

    def __getitem__(self, index) -> T_co:

        # on epoch start (or error), create tf iterator
        if index == 0 or self._iter is None:
            self._iter = iter(self._original_ds_reference)

        # fetch a new batch if current batch is exhausted
        if (
            index % self._internal_cache_size == 0
            or index % self._internal_cache_size >= self._batch_cache_size
        ):
            if self._output_dict:
                batch = self._iter.get_next()
                self._batch_cache = {k: batch[k].numpy() for k in batch.keys()}
                self._batch_cache_size = tf.shape(
                    self._batch_cache[self._first_member]
                )[0].numpy()
            else:
                self._batch_cache = [field.numpy() for field in self._iter.get_next()]
                self._batch_cache_size = tf.shape(self._batch_cache[0])[0].numpy()

        # return as dict or list of numpy arrays
        if self._output_dict:
            return {
                k: self._batch_cache[k][index % self._batch_cache_size]
                for k in self._batch_cache.keys()
            }
        else:
            return [
                field[index % self._batch_cache_size] for field in self._batch_cache
            ]

    def __len__(self):
        return self._len



from decord import VideoReader, AudioReader, cpu
import numpy as np
import os
import sys
import errno
import math

from pathlib import Path
from nova_utils.db_utils import nova_types as nt
from nova_utils.ssi_utils.ssi_stream_utils import Stream
from hcai_datasets.hcai_nova_dynamic.utils.nova_string_utils import (
    merge_role_key,
)
from typing import Union
from abc import ABC, abstractmethod
from enum import Enum


class StrideUnit(Enum):
    MILLISECONDS = (0,)
    SECONDS = (1,)
    FRAMES = 2


class Data(ABC):
    def __init__(
        self,
        role: str = "",
        name: str = "",
        file_ext: str = "stream",
        sr: int = 0,
        data_type: nt.DataTypes = None,
        np_data_type: np.dtype = np.float32,
        is_valid: bool = True,
        sample_data_path: str = "",
        sample_data_shape: tuple = None,
        lazy_loading: bool = False,
    ):
        """
        Abstract base class for all data types
        Args:
            role (str): Role to which the datastream belongs
            name (str): Filename of the datastream without the role
            file_ext (str): File extension of the datastream
            sr (int): Samplerate of the datastream
            data_type (DataTypes): Type of the datastream as specified in the nova_utils package under db_utils.nova_types.novDataTypes
            is_valid (bool): Whether the datastream is valid or not
            sample_data_path (str): A path to an example data file from which all meta information that is not saved in the Nova database can be read
            sample_data_shape (tuple): The default shape of one data sample. Will be overwritten by populate_meta_info if a valid datastream can be found.
            lazy_loading (bool): If set to true only the respective timestamps and filenames will be returned for each window. Not the actual data.
        """
        self.role = role
        self.name = name
        self.is_valid = is_valid
        self.sr = sr
        self.file_ext = file_ext
        self.lazy_loading = lazy_loading
        self.data_type = data_type
        self.np_data_type = np_data_type

        # Set when populate_meta_info is called
        self.sample_data_shape = sample_data_shape
        self.meta_loaded = False
        # self.n_frames_per_window = n_samples_per_window

        # Set when open_file_reader is called
        self.file_path = None
        self.file_reader = None
        self.dur = sys.maxsize

        if sample_data_path:
            if not os.path.isfile(sample_data_path):
                self.meta_loaded = False
                print(
                    f"WARNING: Sample file {sample_data_path} not found. Could not initialize meta data."
                )
                # raise FileNotFoundError( errno.ENOENT, os.strerror(errno.ENOENT), sample_data_path)
            else:
                self.populate_meta_info(sample_data_path)

    def data_stream_opend(self):
        if not self.file_reader:
            print(
                "No datastream opened for {}".format(
                    merge_role_key(self.role, self.name)
                )
            )
            raise RuntimeError("Datastream not loaded")

    def get_info(self):
        if self.meta_loaded:
            if self.lazy_loading:
                return {
                    "frame_start": {"dtype": np.float32, "shape": (1)},
                    "frame_end": {"dtype": np.float32, "shape": (1)},
                    "file_path": {"dtype": np.str, "shape": (None,)},
                }
            return self._get_info_hook()
        else:
            print(
                "Meta data has not been loaded for file {}. Call get_meta_info() first.".format(
                    merge_role_key(self.role, self.name)
                )
            )

    def get_sample(
        self,
        window_start: int,
        window_end: int,
        stride_unit: StrideUnit = StrideUnit.MILLISECONDS,
    ):
        """
        Returns the sample for the respective frames. If lazy loading is true, only the filepath and frame_start, frame_end will be returned.
        """

        if not self.file_reader:
            start_frame, end_frame = sample_window_from_time_interval(
                window_start, window_end, self.sr
            )
            return np.zeros((end_frame - start_frame,) + self.sample_data_shape).astype(self.np_data_type)
        elif self.lazy_loading:
            return {
                "frame_start": window_start,
                "frame_end": window_end,
                "file_path": self.file_path,
            }
        else:
            return self._get_sample_hook(window_start, window_end, stride_unit)

    def open_file_reader(self, path: str):
        """
        Args:
            path ():
        Raises:
            FileNotFoundError: If path is not a file
        """
        self.file_path = path
        if not os.path.isfile(path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)
        if not self.meta_loaded:
            self.populate_meta_info(path)
        self._open_file_reader_hook(path)

    @abstractmethod
    def _get_info_hook(self):
        """
        Returns the features for this datastream to create the DatasetInfo for tensorflow
        """
        ...

    @abstractmethod
    def _get_sample_hook(
        self,
        window_start: int,
        window_end: int,
        stride_unit: StrideUnit = StrideUnit.MILLISECONDS,
    ):
        """
        Returns a data chunk from window start to window end. The window is automatically adjusted to always provide the same number of data samples.

        Args:
            stride_unit ():
        """
        ...

    @abstractmethod
    def _open_file_reader_hook(self, path: str):
        """
        Opens a filereader for the respective datastream. Sets attributes self.file_reader and self.dur
        """
        ...

    @abstractmethod
    def populate_meta_info(self, path: str):
        """
        Opens a data sample from the provided path to extract additional data that is not in the database
        """
        ...

    @abstractmethod
    def close_file_reader(self):
        """
        Closes a filereader for the respective datastream
        """
        ...


class AudioData(Data):
    def __init__(self, **kwargs):
        """

        Args:
            **kwargs ():
        """
        # Overwrite default
        if kwargs.get("sample_data_shape") is None:
            sample_data_shape = (1,)
            kwargs.update({"sample_data_shape": sample_data_shape})
        if kwargs.get("np_data_type") is None:
            kwargs.update({"np_data_type": np.float32})
            kwargs.update({"data_type": nt.DataTypes.AUDIO})

        super().__init__(**kwargs)

    def _get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": np.float32,
        }

    def _get_sample_hook(
        self,
        window_start: int,
        window_end: int,
        stride_unit: StrideUnit = StrideUnit.FRAMES,
    ):

        if stride_unit == StrideUnit.FRAMES:
            window_start_sample = window_start
            window_end_sample = window_end
        else:
            window_start_sample, window_end_sample = sample_window_from_time_interval(
                window_start, window_end, self.sr
            )
        chunk = self.file_reader.get_batch(
            list(range(window_start_sample, window_end_sample))
        ).asnumpy()
        return chunk

    def _open_file_reader_hook(self, path: str):
        self.file_reader = AudioReader(path, ctx=cpu(0), mono=False)
        self.dur = self.file_reader.duration()
        self.n_frames = len(self.file_reader)

    def populate_meta_info(self, path: str):
        """

        Args:
            path ():
        """

        file_reader = AudioReader(path, ctx=cpu(0), mono=False)
        n_channels = file_reader.shape[0]
        self.sample_data_shape = (None, n_channels)
        self.meta_loaded = True

    def close_file_reader(self):
        return True


class VideoData(Data):
    def __init__(self, **kwargs: object) -> object:
        """

        Args:
            **kwargs ():
        """
        # Overwrite default
        if kwargs.get("sample_data_shape") is None:
            sample_data_shape = (480, 640, 3)
            kwargs.update({"sample_data_shape": sample_data_shape})
        if kwargs.get("np_data_type") is None:
            kwargs.update({"np_data_type": np.uint8})
        kwargs.update({"data_type": nt.DataTypes.VIDEO})

        super().__init__(**kwargs)

    def _get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": np.float32,
        }

    def _get_sample_hook(
        self,
        window_start: int,
        window_end: int,
        stride_unit: StrideUnit = StrideUnit.FRAMES,
    ):
        if stride_unit == StrideUnit.FRAMES:
            window_start_sample = window_start
            window_end_sample = window_end
        else:
            window_start_sample, window_end_sample = sample_window_from_time_interval(
                window_start, window_end, self.sr
            )

        chunk = self.file_reader.get_batch(
            list(range(window_start_sample, window_end_sample))
        ).asnumpy()
        return chunk

    def _open_file_reader_hook(self, path: str):
        self.file_reader = VideoReader(path, ctx=cpu(0))
        fps = self.file_reader.get_avg_fps()
        frame_count = len(self.file_reader)
        self.dur = frame_count / fps
        self.n_frames = len(self.file_reader)

    def populate_meta_info(self, path: str):
        file_reader = VideoReader(path)
        self.sample_data_shape = file_reader[0].shape
        self.meta_loaded = True

    def close_file_reader(self):
        return True


class StreamData(Data):
    def __init__(self, **kwargs: object) -> object:
        """

        Args:
            **kwargs ():
        """
        # Overwrite default
        if kwargs.get("sample_data_shape") is None:
            sample_data_shape = (1,)
            kwargs.update({"sample_data_shape": sample_data_shape})
        if kwargs.get("np_data_type") is None:
            kwargs.update({"np_data_type": np.float32})
        kwargs.update({"data_type": nt.DataTypes.FEATURE})

        super().__init__(**kwargs)

    def _get_info_hook(self):
        return merge_role_key(self.role, self.name), {
            "shape": self.sample_data_shape,
            "dtype": self.np_data_type,
        }

    def _get_sample_hook(
        self,
        window_start: int,
        window_end: int,
        stride_unit: StrideUnit = StrideUnit.FRAMES,
    ):
        try:
            self.data_stream_opend()
            if stride_unit == StrideUnit.FRAMES:
                window_start_sample = window_start
                window_end_sample = window_end
            else:
                (
                    window_start_sample,
                    window_end_sample,
                ) = sample_window_from_time_interval(window_start, window_end, self.sr)

            return self.file_reader.data[window_start_sample:window_end_sample]
        except RuntimeError:
            print(
                "Could not get chunk {}-{} from data stream {}".format(
                    window_start, window_end, merge_role_key(self.role, self.name)
                )
            )

    def _open_file_reader_hook(self, path: Path) -> bool:
        stream = Stream(path=path)
        if stream:
            self.file_reader = stream
            self.n_frames = stream.data.shape[0]
            self.dur = self.n_frames / stream.sr
            return True
        else:
            print("Could not open Stream {}".format(str))
            return False

    def close_file_reader(self):
        return True

    def populate_meta_info(self, path: str):
        stream = Stream()
        stream.load_header(path)
        self.sample_data_shape = (stream.dim,)
        self.np_data_type = stream.dtype
        self.meta_loaded = True


##########################
# General helper functions
##########################


# def frame_to_seconds(sr: int, frame: int) -> float:
#    return frame / sr


def seconds_to_sample_nr(time_s: float, sr: float) -> float:
    """
    Calculates the specific sample number in a data stream that corresponds to a given time
    Args:
        sr (float): The sample rate of the data stream
        time_s (float): The timestamp of the sample in seconds

    Returns:
        float: Unrounded index of the sample in the stream for a given time

    """
    return time_s * sr


def milli_seconds_to_sample_nr(time_ms: int, sr: float) -> float:
    """
    Calculates the specific sample number in a data stream that corresponds to a given time
    Args:
        sr (float): The sample rate of the data stream
        time_ms (int): The timestamp of the sample in milliseconds

    Returns:
        float: Unrounded index of the sample in the stream for a given time

    """
    return seconds_to_sample_nr(sr=sr, time_s=time_ms / 1000.0)


def sample_nr_to_seconds(sample_nr: float, sr: float) -> float:
    """
    Calculates the specific time in seconds in a data stream that corresponds to given sample number
    Args:
        sample_nr (float): The unrounded sample number. Might represent a value between to actual samples of the stream.
        sr (float): The average numbers of samples per second.

    Returns:
        float: Unrounded time representation of sample_nr specified in seconds
    """
    return sample_nr * sr


def sample_nr_to_milli_seconds(sample_nr: float, sr: float) -> float:
    """
    Calculates the specific time in milli in a data stream that corresponds to given sample number
    Args:
        sample_nr (float): The unrounded sample number. Might represent a value between to actual samples of the stream.
        sr (float): The average numbers of samples per second.

     Returns:
         float: Unrounded time representation of sample_nr specified in milliseconds
    """
    return sample_nr_to_seconds(sample_nr, sr) * 1000.0


def sample_window_from_time_interval(
    start_time_ms: int, end_time_ms: int, sr: float
) -> tuple[int, int]:
    """
    Calculates the start and end sample number of a sliding window.
    Args:
        start_time_ms (int): Start time of the window in milliseconds
        end_time_ms (int): End time of the window in milliseconds
        sr (float): Sample rate of the data stream

    Returns:
        tuple[int, int]: First and last sample number of the window

    """

    # Get unrounded sample numbers
    start_sample_nr = milli_seconds_to_sample_nr(start_time_ms, sr)
    end_sample_nr = milli_seconds_to_sample_nr(end_time_ms, sr)

    # Rounding start and end to include maximum information
    start_sample_nr = math.floor(start_sample_nr)
    end_sample_nr = math.ceil(end_sample_nr)

    # Assert number of samples in window
    expected_number_of_samples = math.ceil(((end_time_ms - start_time_ms) / 1000) * sr)
    number_of_samples = end_sample_nr - start_sample_nr
    sample_diff = number_of_samples - expected_number_of_samples

    # Fewer then expected samples: We move start_sample number to the left
    # More samples than expected: We move start_sample number to the right
    start_sample_nr += sample_diff

    assert expected_number_of_samples == end_sample_nr - start_sample_nr

    return start_sample_nr, end_sample_nr


def parse_time_string_to_ms(frame: Union[str, int, float]) -> int:
    # if frame is specified milliseconds as string
    if str(frame).endswith("ms"):
        try:
            return int(frame[:-2])
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in milliseconds: {}".format(frame)
            )
    # if frame is specified in seconds as string
    elif str(frame).endswith("s"):
        try:
            frame_s = float(frame[:-1])
            return int(frame_s * 1000)
        except ValueError:
            raise ValueError(
                "Invalid input format for frame in seconds: {}".format(frame)
            )
    # if type is float we assume the input will be seconds
    elif isinstance(frame, float) or "." in str(frame):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is float.".format(
                    frame
                )
            )
            return int(1000 * float(frame))
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))
    # if type is int we assume the input will be milliseconds
    elif isinstance(frame, int) or (isinstance(frame, str) and frame.isdigit()):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is int.".format(
                    frame
                )
            )
            return int(frame)
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))

    print(
        f'WARNING: Could  not automatically parse time "{frame}" to seconds. Returning None '
    )


def parse_time_str(frame: Union[str, int, float]) -> (Union[int, float], StrideUnit):

    if frame is None:
        print(f"WARNING: No timestring to parse. Returning None ")
        return None, StrideUnit.MILLISECONDS

    # if frame is specified milliseconds
    if str(frame).endswith("ms"):
        try:
            return int(frame[:-2]), StrideUnit.MILLISECONDS
        except ValueError:
            raise ValueError(
                "Invalid input format for length specified in milliseconds: {}".format(
                    frame
                )
            )

    # if frame is specified in seconds
    elif str(frame).endswith("s"):
        try:
            return float(frame[:-1]), StrideUnit.SECONDS
        except ValueError:
            raise ValueError(
                "Invalid input format for length specified in seconds: {}".format(frame)
            )

    # if frame is specified in frames
    elif str(frame).endswith("f"):
        try:
            return int(frame[:-1]), StrideUnit.FRAMES
        except ValueError:
            raise ValueError(
                "Invalid input format for length specified in number of frames: {}".format(
                    frame
                )
            )
    # if type is float we assume the input will be seconds
    elif isinstance(frame, float) or "." in str(frame):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is float. Interpreting as seconds".format(
                    frame
                )
            )
            return float(frame), StrideUnit.SECONDS
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))
    # if type is int we assume the input will be milliseconds
    elif isinstance(frame, int) or (isinstance(frame, str) and frame.isdigit()):
        try:
            print(
                "WARNING: Automatically inferred type for frame {} is int. Interpreting as milliseconds".format(
                    frame
                )
            )
            return int(frame), StrideUnit.MILLISECONDS
        except ValueError:
            raise ValueError("Invalid input format for frame: {}".format(frame))

    print(f'WARNING: Could  not automatically parse time "{frame}". Returning None ')
    return None, StrideUnit.MILLISECONDS



from pathlib import Path
from hcai_dataset_utils import import_validator

base_dir = Path(__file__).parent
"""Image Datasets"""
if import_validator.validate_installation(
    base_dir / "hcai_affectnet" / "hcai_affectnet.py"
):
    from hcai_datasets import hcai_affectnet

if import_validator.validate_installation(base_dir / "hcai_ckplus" / "hcai_ckplus.py"):
    from hcai_datasets import hcai_ckplus

if import_validator.validate_installation(base_dir / "hcai_faces" / "hcai_faces.py"):
    from hcai_datasets import hcai_faces

"""Audio Datasets"""
if import_validator.validate_installation(
    base_dir / "hcai_audioset" / "hcai_audioset.py"
):
    from hcai_datasets import hcai_audioset

if import_validator.validate_installation(
    base_dir / "hcai_librispeech" / "hcai_librispeech.py"
):
    from hcai_datasets import hcai_librispeech

if import_validator.validate_installation(
    base_dir / "hcai_is2021_ess" / "hcai_is2021_ess.py"
):
    from hcai_datasets import hcai_is2021_ess

"""Nova Datasets"""
from hcai_datasets import hcai_nova_dynamic



import numpy as np
import pandas as pd
import sys
from numba import njit
from abc import ABC, abstractmethod
from hcai_datasets.hcai_nova_dynamic.utils.nova_string_utils import merge_role_key
from nova_utils.db_utils import nova_types as nt

# TODO: Currently we do not take the rest class into account when calculating the label for the frame. Maybe we should do this
@njit
def _get_overlap(a, start, end):
    """
    Calculating all overlapping intervals between the given array of time intervals and the interval [start, end]
    Args:
        a (): numpy array of shape (1,2), where each entry contains an interval [from, to]
        start (): start of the interval to check
        end (): end of the interval to check

    Returns:
    Numpy array with boolean values. The array is true where the interval specified in a overlaps [start, end]
    """
    annos_for_sample = (
        # annotation is bigger than frame
        ((a[:, 0] <= start) & (a[:, 1] >= end))
        # end of annotation is in frame
        | ((a[:, 1] >= start) & (a[:, 1] <= end))
        # start of annotation is in frame
        | ((a[:, 0] >= start) & (a[:, 0] <= end))
    )
    return annos_for_sample


def _get_anno_majority(a, overlap_idxs, start, end):
    """
    Returns the index of the annotation with the largest overlap with the current frame
    Args:
        a (): numpy array of shape (1,2), where each entry contains an interval [from, to]
        overlap_idxs (): aray of boolean values where a is overlapping the intervall [start, end] (as returned by get _get_overlap())
        start (): start of the interval to check
        end (): end of the interval to check

    Returns:

    """
    # TODO: rewrite for numba jit
    majority_index = -1
    overlap = 0
    for i in np.where(overlap_idxs)[0]:
        if (
            cur_overlap := np.minimum(end, a[i][1]) - np.maximum(start, a[i][0])
        ) > overlap:
            overlap = cur_overlap
            majority_index = i
    return majority_index


def _is_garbage(local_label_id, nova_garbage_label_id):
    # check for nan or compare with garbage label id
    if local_label_id != local_label_id or local_label_id == nova_garbage_label_id:
        return True
    return False


class Annotation(ABC):

    # Uniform label for garbage class handling in python
    GARBAGE_LABEL_ID = np.NAN

    def __init__(self, role: str = "", scheme: str = "", is_valid: bool = True):
        self.role = role
        self.scheme = scheme
        self.is_valid = is_valid

        # Gets set when set_annotation_from_mongo_doc is called
        self.data = None

    @abstractmethod
    def get_info(self):
        """
        Returns the labels for this annotation to create the DatasetInfo for tensorflow
        """
        raise NotImplementedError

    @abstractmethod
    def set_annotation_from_mongo_doc(self, session, time_to_ms=False):
        """
        Returns the labels for this annotation to create the DatasetInfo for tensorflow
        """
        raise NotImplementedError

    @abstractmethod
    def get_label_for_frame(self, start: int, end: int) -> object:
        """
        Returns the label for this frame

        Args:
            start (int): Start time of the frame in milliseconds
            end (int): End time of the frame in milliseconds
        """
        raise NotImplementedError

    @abstractmethod
    def postprocess(self):
        """
        Default post-processing for the respective annotation type
        """
        raise NotImplementedError


class DiscreteAnnotation(Annotation):

    # Class ids and string names as provided from NOVA-DB and required by SSI
    NOVA_REST_CLASS_NAME = "REST"
    NOVA_GARBAGE_LABEL_ID = -1

    # Initialize Rest class id with garbage class id
    REST_LABEL_ID = NOVA_GARBAGE_LABEL_ID

    def __init__(self, labels: dict = None, add_rest_class=False, **kwargs):
        super().__init__(**kwargs)
        self.dataframe = None
        self.type = nt.AnnoTypes.DISCRETE
        self.labels = {
            x["id"]: x["name"] if x["isValid"] else ""
            for x in sorted(labels, key=lambda k: k["id"])
        }
        self.add_rest_class = add_rest_class
        if self.add_rest_class:
            self.REST_LABEL_ID = max(self.labels.keys()) + 1
            self.labels[self.REST_LABEL_ID] = self.NOVA_REST_CLASS_NAME

    def get_info(self):
        return merge_role_key(self.role, self.scheme), {"dtype": np.int32, "shape": 1}

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        self.data = mongo_doc
        if time_to_ms:
            for d in self.data:
                d["from"] = int(float(d["from"]) * 1000)
                d["to"] = int(float(d["to"]) * 1000)

        # Creating Pandas Dataframe version of annotations
        self.dataframe = pd.DataFrame(self.data)

        # Create
        if self.dataframe.empty:
            self.data_interval = np.empty((0, 2), int)
            self.data_values = np.empty((0, 2), int)
        else:
            # Creating numpy array of annotations for fast access
            # Splitting the annotations into interval and data array
            self.data_interval = self.dataframe[["from", "to"]].values.astype(int)
            self.data_values = self.dataframe[["id", "conf"]].values

    def get_label_for_frame_legacy(self, start, end):

        # If we don't have any data we return the garbage label
        if self.data == -1:
            return -1

        else:
            # Finding all annos that overlap with the frame
            def is_overlapping(af, at, ff, ft):

                # anno is larger than frame
                altf = af <= ff and at >= ft

                # anno overlaps frame start
                aofs = at >= ff and at <= ft

                # anno overlaps frame end
                aofe = af >= ff and af <= ft

                return altf or aofs or aofe

            annos_for_sample = list(
                filter(
                    lambda x: is_overlapping(x["from"], x["to"], start, end), self.data
                )
            )

            # No label matches
            if not annos_for_sample:
                if self.add_rest_class:
                    return len(self.labels.values()) - 1
                else:
                    return -1

            majority_sample_idx = np.argmax(
                list(
                    map(
                        lambda x: min(end, x["to"]) - max(start, x["from"]),
                        annos_for_sample,
                    )
                )
            )

            return annos_for_sample[majority_sample_idx]["id"]

    def get_label_for_frame_np(self, start, end):

        overlap_idxs = _get_overlap(self.data_interval, start, end)

        # If no label overlaps the requested frame we return rest class. If add_rest_class = False garbage label will be returned instead
        if not overlap_idxs.any():
            return self.REST_LABEL_ID

        majority_idx = _get_anno_majority(self.data_interval, overlap_idxs, start, end)
        label = self.data_values[majority_idx, 0]
        if _is_garbage(label, self.NOVA_GARBAGE_LABEL_ID):
            return Annotation.GARBAGE_LABEL_ID
        return label

    def get_label_for_frame(self, start, end):
        return self.get_label_for_frame_np(start, end)

    # TODO: postprocessing MINGAP -> MINDUR -> MINGAP (filter -> pack -> filter)
    def postprocess(self):
        pass


class FreeAnnotation(Annotation):
    """
    The FREE annotation scheme is used for any form of free text.
    """

    def __init__(self, **kwargs):
        self.dataframe = None
        self.type = nt.AnnoTypes.FREE
        super().__init__(**kwargs)

    def get_info(self):
        return merge_role_key(self.role, self.scheme), {
            "dtype": np.str,
            "shape": (None,),
        }

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        self.data = mongo_doc
        if time_to_ms:
            for d in self.data:
                d["from"] = int(d["from"] * 1000)
                d["to"] = int(d["to"] * 1000)

        # Creating Pandas Dataframe version of annotations
        self.dataframe = pd.DataFrame(self.data)

        # Create
        if self.dataframe.empty:
            self.data_interval = np.empty((0, 2), int)
            self.data_values = np.empty((0, 2), int)
        else:
            # Creating numpy array of annotations for fast access
            # Splitting the annotations into interval and data array
            self.data_interval = self.dataframe[["from", "to"]].values.astype(int)
            self.data_values = self.dataframe[["name", "conf"]].values

    def get_label_for_frame_legacy(self, start, end):

        # If we don't have any data we return the garbage label
        if self.data == -1:
            return -1

        else:
            # Finding all annos that overlap with the frame
            def is_overlapping(af, at, ff, ft):

                # anno is larger than frame
                altf = af <= ff and at >= ft

                # anno overlaps frame start
                aofs = at >= ff and at <= ft

                # anno overlaps frame end
                aofe = af >= ff and af <= ft

                return altf or aofs or aofe

            annos_for_sample = list(
                filter(
                    lambda x: is_overlapping(x["from"], x["to"], start, end), self.data
                )
            )

            # No label matches
            if not annos_for_sample:
                return [""]

            return [a["name"] for a in annos_for_sample]

    def get_label_for_frame_np(self, start, end):
        annos_for_sample = _get_overlap(self.data_interval, start, end)

        # No label matches
        if not annos_for_sample.any():
            return [""]

        return self.data_values[annos_for_sample, 0]

    def get_label_for_frame(self, start, end):
        return self.get_label_for_frame_np(start, end)

    def postprocess(self):
        pass


class ContinuousAnnotation(Annotation):

    # Class ids and string names as provided from NOVA-DB and required by SSI
    NOVA_GARBAGE_LABEL_VALUE = np.NAN

    MISSING_DATA_LABEL_VALUE = sys.float_info.min

    def __init__(self, sr=0, min_val=0, max_val=0, **kwargs):
        super().__init__(**kwargs)
        self.type = nt.AnnoTypes.CONTINUOUS
        self.sr = sr
        self.min_val = min_val
        self.max_val = max_val
        self.labels = {1: self.scheme}

    def get_info(self):
        return merge_role_key(self.role, self.scheme), {"dtype": np.int32, "shape": 1}

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        # Numpy array with shape (len_data, 2) where the second dimension is a respective tuple (confidence, score)
        self.data = np.array([(i["score"], i["conf"]) for i in mongo_doc])

    def get_label_for_frame(self, start, end):
        # returns zero if session duration is longer then labels
        s = int(start * self.sr / 1000)
        e = int(end * self.sr / 1000)

        if len(self.data) >= e:
            frame = self.data[s:e]
            frame_data = frame[:, 0]
            frame_conf = frame[:, 1]
        else:
            return self.MISSING_DATA_LABEL_VALUE

        # TODO: Return timeseries instead of average
        conf = sum(frame_conf) / max(len(frame_conf), 1)
        label = sum(frame_data) / max(len(frame_data), 1)

        # If frame evaluates to garbage label discard sample
        if _is_garbage(label, self.NOVA_GARBAGE_LABEL_VALUE):
            return self.NOVA_GARBAGE_LABEL_VALUE
        else:
            return label

    def postprocess(self):
        pass


class DiscretePolygonAnnotation(Annotation):
    def __init__(self, labels={}, sr=0, **kwargs):
        super().__init__(**kwargs)
        self.type = nt.AnnoTypes.DISCRETE_POLYGON

        self.labels = {
            str(x["id"]): (x["name"], x["color"]) if x["isValid"] else ""
            for x in sorted(labels, key=lambda k: k["id"])
        }
        self.sr = sr

    # TODO MARCO type wird so für tensorflow nicht funktionieren
    def get_info(self):
        return merge_role_key(self.role, self.scheme), {
            "dtype": np.float64,
            "shape": (None, 2),
        }

    def set_annotation_from_mongo_doc(self, mongo_doc, time_to_ms=False):
        self.data = mongo_doc

    def get_label_for_frame(self, start, end):
        # return the last frame
        frame_nr = int((end / 1000) * self.sr)
        if len(self.data) > frame_nr:
            return self.data[frame_nr - 1]
        else:
            return -1

    def postprocess(self):
        pass



"""hcai_faces dataset."""

import tensorflow as tf
import tensorflow_datasets as tfds
import pandas as pd
from tensorflow_datasets.core.splits import Split
from pathlib import Path
from hcai_dataset_utils.statistics import Statistics
from hcai_datasets.hcai_faces.hcai_faces_iterable import HcaiFacesIterable

_DESCRIPTION = """
FACES is a set of images of naturalistic faces of 171 young (n = 58), middle-aged (n = 56), and older (n = 57) women and 
men displaying each of six facial expressions: neutrality, sadness, disgust, fear, anger, and happiness. 
The FACES database was developed between 2005 and 2007 by Natalie C. Ebner, Michaela Riediger, 
and Ulman Lindenberger at the Center for Lifespan Psychology, Max Planck Institute for Human Development, Berlin, Germany.
"""

_CITATION = """
@article{ebner2010faces,
  title={FACES—A database of facial expressions in young, middle-aged, and older women and men: Development and validation},
  author={Ebner, Natalie C and Riediger, Michaela and Lindenberger, Ulman},
  journal={Behavior research methods},
  volume={42},
  number={1},
  pages={351--362},
  year={2010},
  publisher={Springer}
}
"""


class HcaiFaces(tfds.core.GeneratorBasedBuilder, HcaiFacesIterable, Statistics):
    """DatasetBuilder for hcai_faces dataset."""

    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }

    def __init__(self, dataset_dir, *args, **kwargs):
        HcaiFacesIterable.__init__(self, *args, dataset_dir=dataset_dir, **kwargs)
        tfds.core.GeneratorBasedBuilder.__init__(self, *args, **kwargs)

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            metadata=tfds.core.MetadataDict({}),
            features=tfds.features.FeaturesDict(
                {
                    # These are the features of your dataset like images, labels ...
                    "index": tf.string,
                    "image": tfds.features.Image(shape=(3543, 2835, 3)),
                    "id": tf.int64,
                    "age": tfds.features.ClassLabel(names=["y", "o", "m"]),
                    "gender": tfds.features.ClassLabel(names=["m", "f"]),
                    "emotion": tfds.features.ClassLabel(
                        names=["a", "d", "f", "h", "n", "s"]
                    ),
                    "set": tfds.features.ClassLabel(names=["a", "b"]),
                    "rel_file_path": tf.string,
                }
            ),
            supervised_keys=("image", "emotion"),
            homepage="https://dataset-homepage/",
            citation=_CITATION,
        )

    def _populate_meta_data(self, data):
        df = pd.DataFrame(data, columns=["id", "age", "gender", "emotion", "set"])
        self._populate_stats(df)

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):
        """Returns SplitGenerators."""
        self._populate_meta_data([data for file, data in self.parsed])

        return {
            Split.TRAIN: self._generate_examples(self.parsed),
        }

    def _generate_examples(self, files):
        for sample in self.yield_samples(files):
            yield sample["index"], sample



import os
import errno
import copy
import warnings
import enum

from configparser import ConfigParser
from typing import Union
from datetime import datetime
from pymongo import MongoClient
from pymongo.results import InsertOneResult, UpdateResult


class NovaDBHandler:
    ANNOTATOR_COLLECTION = "Annotators"
    SCHEME_COLLECTION = "Schemes"
    STREAM_COLLECTION = "Streams"
    ROLE_COLLECTION = "Roles"
    ANNOTATION_COLLECTION = "Annotations"
    SESSION_COLLECTION = "Sessions"
    ANNOTATION_DATA_COLLECTION = "AnnotationData"

    def __init__(self, db_config_path=None, db_config_dict=None):

        # Connecting to the database
        if db_config_path:
            if os.path.isfile(db_config_path):
                cfg = self.read_config(db_config_path)
                self.ip = str(cfg["DB"]["ip"])
                self.port = int(cfg["DB"]["port"])
                self.user = str(cfg["DB"]["user"])
                self.password = str(cfg["DB"]["password"])
            else:
                raise FileNotFoundError(
                    "No database config file found at {}".format(db_config_path)
                )

        # If a db config_dict is specified overwrite the config from path
        if db_config_dict:
            if db_config_path:
                print(
                    "WARNING! database config are specifed as file AND and as dictionary. Using the dictionary."
                )
            self.ip = db_config_dict["ip"]
            self.port = db_config_dict["port"]
            self.user = db_config_dict["user"]
            self.password = db_config_dict["password"]

        if not (self.ip or self.port or self.user or self.password):
            print(
                "WARNING! No valid nova database config found for path {} and dict {} \n Found config parameters are ip:{}, port{}, user: {}. Also check your password.".format(
                    db_config_path, db_config_dict, self.ip, self.port, self.user
                )
            )

        self.client = MongoClient(
            host=self.ip, port=self.port, username=self.user, password=self.password
        )

    def print_config(self, cfg: ConfigParser, cfg_path: str) -> None:
        """
        Prints the provided configuration on the console
        Args:
            cfg (ConfigParser): The main configuration parser
            cfg_path (str): The file to load the configuration from
        """
        print("Loaded config from {}:".format(cfg_path))
        print("---------------------")
        for sec_name, sec_dict in cfg._sections.items():
            print(sec_name)
            for k, v in sec_dict.items():
                if k == "password":
                    continue
                else:
                    print("\t{} : {}".format(k, v))
        print("---------------------")

    def read_config(self, cfg_path: str) -> ConfigParser:
        """
        Reads a NovaServer configuration from a file
        Args:
            cfg_path (str): The file to load the configuration from

        Returns:
            ConfigParser: The parsed configuration

        """
        config = ConfigParser()
        config.read(cfg_path)
        self.print_config(config, cfg_path)
        return config

    # Reading from database
    def get_docs_by_prop(
        self, vals: Union[list[str], str], property: str, database: str, collection: str
    ) -> list[dict]:
        """
        Fetching a document from the mongo db collection in the respective database where the passed values are matching a specific property in the collection.
        Args:
            vals (Union[list[str], str]): Value(s) of a property in the document. Might be a single value or list of values.
            property (str): Property to look for the passed values
            database (str): Name of the database to search
            collection (str): Name of the collection within the database to search

        Returns:
            list[dict]: List of documents that match the specified criteria
        """
        filter = []

        if not type(vals) == type(list()):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}
        ret = list(self.client[database][collection].find(filter))
        return ret

    def get_schemes(self, dataset: str, schemes: list) -> list[dict]:
        """
        Retreives all schemes matchin the provided criteria
        Args:
            dataset (str): Name of the dataset. Must match the respective entry in the MongoDB
            schemes (str): List of schemes for which the mongo database entries are fetched

        Returns:
            list[dict]: List of schemes that match the specified criteria
        """

        if not schemes:
            print("WARNING: No Schemes have been requested. Returning empty list.")
            return []

        # if not dataset in self.datasets:
        # raise ValueError('{} not found in datasets'.format(dataset))

        mongo_schemes = []
        for scheme in schemes:
            mongo_scheme = self.get_docs_by_prop(
                scheme, "name", dataset, self.SCHEME_COLLECTION
            )
            if not mongo_scheme:
                print(
                    f"WARNING: No scheme {scheme} found in database for dataset {dataset}"
                )
            else:
                mongo_schemes.append(mongo_scheme[0])

        if not mongo_schemes:
            raise ValueError(
                "No entries for schemes {} found in database".format(schemes)
            )

        return mongo_schemes

    def get_session_info(self, dataset: str, session: Union[list, str]) -> list[dict]:
        """
        Fetches the session object that matches the specified criteria from the nova database and returns them as a python readable dictionary.

        Args:
          dataset (str): Name of the dataset. Must match the respective entry in the MongoDB
          session (Union[list, str]): Session or list of sessions for which the mongo database entries are fetched

        Returns:
          list[dict]: List of sessions that match the specified criteria
        """
        mongo_session = self.get_docs_by_prop(
            session, "name", dataset, self.SESSION_COLLECTION
        )
        return mongo_session

    def get_data_streams(self, dataset, data_streams):
        """
        Fetches the stream objects that matches the specified criteria from the nova database and returns them as a python readable dictionary.
        Args:
          dataset:
          session:
          role_list:
          data_stream_list:
        """
        # if not dataset in self.datasets:
        #  raise ValueError('{} not found in datasets'.format(dataset))

        if not data_streams:
            print("WARNING: No Datastreams have been requested. Returning empty list.")
            return []

        mongo_streams = []
        for stream in data_streams:
            mongo_stream = self.get_docs_by_prop(
                stream, "name", dataset, self.STREAM_COLLECTION
            )
            if not mongo_stream:
                print("WARNING: No stream {} found in database".format(stream))
            else:
                mongo_streams.append(mongo_stream[0])

        if not mongo_streams:
            raise ValueError("no entries for datastream {} found".format(data_streams))

        return mongo_streams

    def get_annotation_docs(
        self,
        mongo_schemes,
        mongo_sessions,
        mongo_annotators,
        mongo_roles,
        database,
        collection,
    ):
        """
        Fetches all annotation objects that match the specified criteria from the nova database
        Args:
          mongo_schemes:
          mongo_sessions:
          mongo_annotators:
          mongo_roles:
          database:
          collection:
          client:

        Returns:

        """
        scheme_filter = []
        role_filter = []
        annotator_filter = []
        session_filter = []

        for ms in mongo_schemes:
            scheme_filter.append({"scheme_id": ms["_id"]})

        for mse in mongo_sessions:
            session_filter.append({"session_id": mse["_id"]})

        for ma in mongo_annotators:
            annotator_filter.append({"annotator_id": ma["_id"]})

        for mr in mongo_roles:
            role_filter.append({"role_id": mr["_id"]})

        filter = {
            "$and": [
                {"$or": scheme_filter},
                {"$or": session_filter},
                {"$or": role_filter},
                {"$or": annotator_filter},
            ]
        }

        ret = list(self.client[database][collection].find(filter))
        return ret

    def get_annos(
        self,
        dataset: str,
        scheme: str,
        session: str,
        annotator: str,
        roles: Union[list, str],
    ) -> list:
        """
        Fetches all annotations that matches the specified criteria from the nova database and returns them as a list of python readable dictionaries.
        Args:
          dataset:
          scheme:
          session:
          annotator:
          roles:

        Returns:

        """
        mongo_schemes = self.get_docs_by_prop(
            scheme, "name", dataset, self.SCHEME_COLLECTION
        )
        if not mongo_schemes:
            warnings.warn(f"Unknown scheme {scheme} found")
            return []
        mongo_annotators = self.get_docs_by_prop(
            annotator, "name", dataset, self.ANNOTATOR_COLLECTION
        )
        if not mongo_annotators:
            warnings.warn(f"Unknown annotator {annotator} found")
            return []
        mongo_roles = self.get_docs_by_prop(
            roles, "name", dataset, self.ROLE_COLLECTION
        )
        if not mongo_roles:
            warnings.warn(f"Unknown role {roles} found")
            return []
        mongo_sessions = self.get_docs_by_prop(
            session, "name", dataset, self.SESSION_COLLECTION
        )
        if not mongo_sessions:
            warnings.warn(f"Unknown for session {session} found")
            return []

        mongo_annos = self.get_annotation_docs(
            mongo_schemes,
            mongo_sessions,
            mongo_annotators,
            mongo_roles,
            dataset,
            self.ANNOTATION_COLLECTION,
        )

        # getting the annotation data and the session name
        if not mongo_annos:
            raise FileNotFoundError(
                errno.ENOENT,
                "No such annotation",
                f"annotator: {annotator} - scheme: {scheme} - session: {session} - role: {roles}",
            )

        else:
            # TODO: adapt for multiple roles, annotators etc.
            label = self.get_data_docs_by_prop(
                mongo_annos[0]["data_id"], "_id", dataset
            )
            label = label["labels"]

        return label

    def insert_doc_by_prop(
        self, doc: dict, database: str, collection: str
    ) -> InsertOneResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].insert_one(doc)
        return ret

    def update_doc_by_prop(
        self, doc: dict, database: str, collection: str
    ) -> UpdateResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].update_one(
            {"_id": doc["_id"]}, {"$set": doc}
        )
        return ret

    def update_doc_by_id(
        self, _id: str, doc: dict, database: str, collection: str
    ) -> UpdateResult:
        """
        Uploading a document to the database using the specificed parameters
        Args:
          _id: ID of doc, which has to be updated
          docs: List of dictionaries with objects to upload to the database
          database: The name of the database to search
          collection: The name of the collection within the database to search

        Returns:
          str: ObjectID of the inserted objects or an empty list in case of failure
        """
        ret = self.client[database][collection].update_one({"_id": _id}, {"$set": doc})
        return ret

    # TODO: Remove Restclass Labels in discrete Cases
    # TODO: Consider "forced overwrite"
    # TODO: Add Backup case
    # TODO: Call preprocess of annotation
    def set_annos(
        self,
        database: str,
        scheme: str,
        session: str,
        annotator: str,
        role: str,
        annos: list,
        is_finished: bool = False,
        is_locked: bool = False,
    ) -> str:
        """
        Uploading annotations to the database
        Args:
          database:
          scheme:
          session:
          annotator:
          role:
          annos:

        Returns: Object ID of the inserted annotations. Empty string in case of failure
        """
        mongo_scheme = self.get_mongo_scheme(scheme, database)
        mongo_annotator = self.get_mongo_annotator(annotator, database)
        mongo_role = self.get_mongo_role(role, database)
        mongo_session = self.get_mongo_session(session, database)

        # Check if annotations already exist
        mongo_annos = self.get_annotation_docs(
            mongo_scheme,
            mongo_session,
            mongo_annotator,
            mongo_role,
            database,
            self.ANNOTATION_COLLECTION,
        )

        # Check for existing annotations
        mongo_anno_id = None
        mongo_data_id = None
        if mongo_annos:
            if mongo_annos[0]["isLocked"]:
                warnings.warn(
                    f"Can't overwrite locked annotation {str(mongo_annos[0]['_id'])}"
                )
                return ""
            else:
                warnings.warn(
                    f"Overwriting existing annotation {str(mongo_annos[0]['_id'])}"
                )
                mongo_anno_id = mongo_annos[0]["_id"]
                mongo_data_id = mongo_annos[0]["data_id"]

        # Upload label data
        mongo_label_doc = {"labels": annos}
        if mongo_data_id:
            mongo_label_doc["_id"] = mongo_data_id
            success = self.update_doc_by_prop(
                doc=mongo_label_doc,
                database=database,
                collection=self.ANNOTATION_DATA_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unknown error update database entries for Annotation data {mongo_data_id}"
                )
                return ""
            else:
                data_id = mongo_data_id
        else:
            success = self.insert_doc_by_prop(
                doc=mongo_label_doc,
                database=database,
                collection=self.ANNOTATION_DATA_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unexpected error uploading annotation data for {database} - {session} - {scheme} - "
                    f"{annotator}. Upload failed."
                )
                return ""
            else:
                data_id = success.inserted_id

        # Upload annotation information
        mongo_anno_doc = {
            "data_id": data_id,
            "annotator_id": mongo_annotator[0]["_id"],
            "role_id": mongo_role[0]["_id"],
            "scheme_id": mongo_scheme[0]["_id"],
            "session_id": mongo_session[0]["_id"],
            "isFinished": is_finished,
            "isLocked": is_locked,
            "date": datetime.today().replace(microsecond=0),
        }

        if mongo_anno_id:
            mongo_anno_doc["_id"] = mongo_anno_id
            success = self.update_doc_by_prop(
                doc=mongo_anno_doc,
                database=database,
                collection=self.ANNOTATION_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unexpected error uploading annotations for {database} - {session} - {scheme} - {annotator}. Upload failed."
                )
                return ""
            else:
                anno_id = mongo_anno_id
        else:
            success = self.insert_doc_by_prop(
                doc=mongo_anno_doc,
                database=database,
                collection=self.ANNOTATION_COLLECTION,
            )
            if not success.acknowledged:
                warnings.warn(
                    f"Unexpected error uploading annotations for {database} - {session} - {scheme} - {annotator}. Upload failed."
                )
                return ""
            else:
                anno_id = success.inserted_id
        return anno_id

    def set_data_streams(
        self,
        database: str,
        file_name: str,
        file_ext: str,
        stream_type: str,
        is_valid: bool,
        sr: float,
        dimlabels: list,
        overwrite: bool = False,
    ):

        # build doc
        mongo_stream_doc = {
            "name": file_name,
            "fileExt": file_ext,
            "type": stream_type,
            "isValid": is_valid,
            "sr": sr,
            "dimlabels": dimlabels,
        }

        mongo_stream = self.get_docs_by_prop(
            file_name, "name", database, self.STREAM_COLLECTION
        )

        if mongo_stream:

            # check if datastream already exists
            if overwrite:
                mongo_stream_doc["_id"] = mongo_stream[0]["_id"]
                success = self.update_doc_by_prop(
                    doc=mongo_stream_doc,
                    database=database,
                    collection=self.STREAM_COLLECTION,
                )
            else:
                print(
                    f"INFO: Stream {file_name} already exists in database. Skip adding stream."
                )
                return
        else:
            # insert datastream
            success = self.insert_doc_by_prop(
                doc=mongo_stream_doc,
                database=database,
                collection=self.STREAM_COLLECTION,
            )
            if success.acknowledged:
                mongo_stream_doc["_id"] = success.inserted_id

        if not success.acknowledged:
            warnings.warn(
                f"Unexpected error adding stream for {database} - {file_name}.{file_ext}. Upload failed."
            )
            return ""


        return mongo_stream_doc["_id"]

    def get_mongo_scheme(self, scheme, database):
        mongo_scheme = self.get_docs_by_prop(
            scheme, "name", database, self.SCHEME_COLLECTION
        )
        if not mongo_scheme:
            warnings.warn(f"Unknown scheme {scheme} found")
            return ""

        return mongo_scheme

    def get_mongo_annotator(self, annotator, database):
        mongo_annotator = self.get_docs_by_prop(
            annotator, "name", database, self.ANNOTATOR_COLLECTION
        )
        if not mongo_annotator:
            warnings.warn(f"Unknown annotator {annotator} found")
            return ""

        return mongo_annotator

    def get_mongo_role(self, role, database):
        mongo_role = self.get_docs_by_prop(role, "name", database, self.ROLE_COLLECTION)
        if not mongo_role:
            warnings.warn(f"Unknown role {role} found")
            return ""

        return mongo_role

    def get_mongo_session(self, session, database):
        mongo_session = self.get_docs_by_prop(
            session, "name", database, self.SESSION_COLLECTION
        )
        if not mongo_session:
            warnings.warn(f"Unknown for session {session} found")
            return ""

        return mongo_session

    def delete_doc_with_tail(self, doc_id_to_remove, database):
        while doc_id_to_remove is not None:
            remove_id = copy.deepcopy(doc_id_to_remove)
            result = self.get_fields_by_properties(
                doc_id_to_remove,
                "_id",
                "nextEntry",
                database,
                self.ANNOTATION_DATA_COLLECTION,
            )
            if result is not None and "nextEntry" in result:
                doc_id_to_remove = result["nextEntry"]
            else:
                doc_id_to_remove = None

            self.delete_doc_by_prop(
                remove_id, "_id", database, self.ANNOTATION_DATA_COLLECTION
            )

    def delete_doc_by_prop(
        self, vals: Union[list, str], property: str, database: str, collection
    ):
        filter = []

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$and": filter}

        return self.client[database][collection].delete_one(filter)

    def get_data_docs_by_prop(
        self, vals: Union[list, str], property: str, database: str
    ):
        filter = []

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}

        result = list(
            self.client[database][self.ANNOTATION_DATA_COLLECTION].find(filter)
        )[0]
        if "nextEntry" in result:
            return self.merge_collections(result, database)

        return result

    def get_fields_by_properties(
        self,
        vals: Union[list, str],
        property,
        fields: Union[list, str],
        database: str,
        collection: str,
    ):
        filter = []
        fields_dict = {}

        if not isinstance(fields, list):
            fields = [fields]

        for n in fields:
            fields_dict[n] = 1

        if not isinstance(vals, list):
            vals = [vals]

        for n in vals:
            filter.append({property: n})

        filter = {"$or": filter}

        return self.client[database][collection].find_one(filter, fields_dict)

    def merge_collections(self, doc, database):
        next_id = doc["nextEntry"]
        new_doc = self.get_data_docs_by_prop(next_id, "_id", database)
        doc["labels"] += new_doc["labels"]

        return doc


if __name__ == "__main__":
    db_handler = NovaDBHandler("../../local/nova_db_test.cfg")

    test_cont = False
    test_cat = False
    test_free = True

    # Test continuous data download and upload
    if test_cont:
        dataset = "aria-noxi"
        session = "004_2016-03-18_Paris"
        scheme = "engagement"
        annotator = "system"
        roles = ["novice"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

    # Test categorical data download and upload
    if test_cat:
        dataset = "roxi"
        session = "001"
        scheme = "emotionalbursts"
        annotator = "gold"
        roles = ["player1"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

        new_annotator = "test"
        new_annos = [
            {"from": 0, "to": 10, "id": 1, "conf": 0.5},
            {"from": 20, "to": 25, "id": 1, "conf": 1},
            {"from": 30, "to": 35, "id": 1, "conf": 1},
        ]

        db_handler.set_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=new_annotator,
            role=roles[0],
            annos=new_annos,
        )

    # Test free label download and upload
    if test_free:
        dataset = "kassel_therapie_korpus"
        session = "OPD_102"
        scheme = "transcript"
        annotator = "system"
        roles = ["therapist"]

        mongo_scheme = db_handler.get_schemes(dataset=dataset, schemes=[scheme])
        annos = db_handler.get_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=annotator,
            roles=roles,
        )

        new_annotator = "schildom"
        new_annos = [
            {"from": 0, "to": 10, "conf": 1, "name": "das"},
            {"from": 20, "to": 25, "conf": 1, "name": "geht"},
            {"from": 30, "to": 35, "conf": 1, "name": "ja"},
        ]

        db_handler.set_annos(
            dataset=dataset,
            scheme=scheme,
            session=session,
            annotator=new_annotator,
            role=roles[0],
            annos=new_annos,
        )

    print("Done")
